package cs108;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;

import org.junit.jupiter.api.Test;

public abstract class SSetTest {
    public abstract <E> SSet<E> newSet();
    
    @Test
    public void sizeIsInitially0() {
        assertEquals(0, newSet().size());
    }
    
    @Test
    public void sizeIncreasesWhenAdding() {
        SSet<Integer> l = newSet();
        for (int i = 0; i < 10; ++i) {
            assertEquals(i, l.size());
            l.add(i);
        }
    }

    @Test
    public void sizeDecreasesWhenRemoving() {
        SSet<Integer> l = newSet();
        for (int i = 1; i <= 10; ++i)
            l.add(i);
        for (int s = 10; s > 0; --s) {
            assertEquals(s, l.size());
            l.remove(s);
        }
        assertEquals(0, l.size());
    }

    @Test
    public void toStringWorksOnEmptySet() {
        assertEquals("{}", newSet().toString());
    }
    
    private enum Action { ADD, REMOVE, CONTAINS }

    @Test
    public void behavesLikeJavaList() {
        Random rng = new Random();
        int maxValue = 100;
        
        Set<Integer> js = new HashSet<>();
        SSet<Integer> s = newSet();
        
        for (int i = 1; i < 100000; ++i) {
            Action action = js.isEmpty()
                    ? Action.ADD
                            : Action.values()[rng.nextInt(Action.values().length)];
            switch (action) {
            case ADD: {
                int e = rng.nextInt(maxValue);
                js.add(e);
                s.add(e);
            } break;
            
            case REMOVE: {
                int j = rng.nextInt(maxValue);
                js.remove(j);
                s.remove(j);
            } break;

            case CONTAINS: {
                int j = rng.nextInt(maxValue);
                boolean jc = js.contains(j);
                boolean c = s.contains(j);
                assertEquals(jc, c);
            } break;
            
            default:
                throw new Error();
            }
            
            assertEquals(js.size(), s.size());
            assertPairwiseEqual(js, s);
        }

        Iterator<Integer> ji = js.iterator();
        Iterator<Integer> i = s.iterator();
        while (ji.hasNext()) {
            assertTrue(i.hasNext());
            int je = ji.next(), e = i.next();
            if (je == e) {
                ji.remove();
                i.remove();
            }
        }
        assertPairwiseEqual(js, s);
    }
    
    private static <E> void assertPairwiseEqual(Set<E> js, SSet<E> s) {
        for (E e: js)
            assertTrue(s.contains(e));
        for (E e: s)
            assertTrue(js.contains(e));
    }
}
